/**
 *  OdetteJ - Open Source Odette FTP implementation in Java(TM).
 *  Copyright (C) 2005 Fossil E-Commerce, http://www.fossilec.com
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  ------------------------------------------------------------------------
 *  $Id: SpeakerState.java,v 1.10 2005/09/10 23:58:11 rmarins Exp $
 *  ------------------------------------------------------------------------
 */

package org.fossilec.odettej;

import java.nio.ByteBuffer;
import java.util.Date;
import java.util.Iterator;

import org.fossilec.odettej.service.AnswerReason;
import org.fossilec.odettej.service.CommandExchangeBuffer;
import org.fossilec.odettej.service.CommandIdentifier;
import org.fossilec.odettej.service.EndSessionReason;
import org.fossilec.odettej.service.DataExchangeBuffer;
import org.fossilec.odettej.service.OdetteExchangeBuffer;
import org.fossilec.odettej.service.RecordFormat;
import org.fossilec.odettej.service.TransferMode;

final class SpeakerState extends AbstractState {

	public SpeakerState(Session context) {
		super(context);
	}

	protected void listen() throws OdetteFTPException {
		illegalState();
	}

	public void startSession() throws OdetteFTPException {

		Session context = getContext();
		PreferencesUtil prefs = getPreferences();

		String userCode = prefs.getLocalUser();
		String password = prefs.getLocalPassword();
		TransferMode mode = prefs.getTransferMode();
		boolean restart = prefs.isRestartSupported();

		/* Begin Start Session Phase */
		waitForReadyMessage();

		negotiateStartSession(userCode, password, mode, restart);

		/*
		 * As the initiator will always be the Speaker, when is specified the
		 * receiver only just change direction after establishing a session.
		 */
		if (prefs.getTransferMode() == TransferMode.RECEIVER_ONLY) {
			context.state().changeDirection();
		}
	}

	public void startFile(VirtualFile file, String destination,
			String originator, String userData, String reserved)
			throws OdetteFTPException {

		Session context = getContext();

		/**
		 * Sending Start File command.
		 */
		CommandExchangeBuffer sfid;
		sfid = CommandExchangeBuffer.startFile(file.getDatasetName(), "", file
				.getDateTime(), userData, destination, originator, file
				.getFormat(), file.getRecordSize(), file.getFileSize(), file
				.getOffset());

		send(sfid);

		/**
		 * Receiving the Start File Answer command.
		 */
		CommandExchangeBuffer answer = (CommandExchangeBuffer) receive();

		if (answer.getIdentifier() == CommandIdentifier.SFPA) {
			int answerCount = Integer.parseInt(answer.getParameter("SFPAACNT"));
			if ((answerCount < file.getOffset()) && (answerCount >= 0)) {
				/* reset restart offset with the new value */
				file.setupRestart(answerCount);

			} else if (answerCount > file.getOffset()) {
				abnormalRelease(EndSessionReason.PROTOCOL_VIOLATION,
						"Returned restart position greater than specified.");
			}

			/* Start File Positive Answer acts as a implicit Credit command */
			resetCredits();

		} else if (answer.getIdentifier() == CommandIdentifier.SFNA) {
			/* Start File Negative Answer */
			CommandExchangeBuffer command = (CommandExchangeBuffer) answer;

			String code = command.getParameter("SFNAREAS");
			AnswerReason reason = (AnswerReason) AnswerReason.parse(code);
			throw new FileTransferException(reason, "Start File Negative "
					+ "Answer: " + code);

		} else if (answer.getIdentifier() != CommandIdentifier.SFPA) {
			/* Unexpected command */
			abnormalRelease(EndSessionReason.PROTOCOL_VIOLATION,
					"SFPA or SFNA commands were expected");
		}

		context.setFile(file);
	}

	public void dataRegime() throws OdetteFTPException {

		Session context = getContext();

		PreferencesUtil prefs = getPreferences();
		boolean compression = prefs.isCompressionSupported();
		int sdeb = prefs.getBufferSize();

		VirtualFile file = context.getFile();

		/**
		 * Data Exchange phase.
		 */

		Iterator it = file.getRecordIterator();
		DataExchangeBuffer deb = new DataExchangeBuffer(sdeb);

		boolean lastRecord;
		ByteBuffer record;

		while (it.hasNext()) {

			record = (ByteBuffer) it.next();
			lastRecord = it.hasNext();

			while (record.hasRemaining()) {

				deb.fill(record, compression, file.getFormat(), lastRecord);

				if ((!deb.hasRemaining()) || (!it.hasNext())) {
					deb.prepareSubmit();
					sendData(deb);

					deb.clear();
				}

			}
		}

	}

	private void sendData(DataExchangeBuffer data) throws OdetteFTPException {

		try {
			Thread.sleep(1000);
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		send(data);

		consumeCredit();

		if (!hasCredits()) {
			waitSetCredit();
			resetCredits();
		}
	}

	private void waitSetCredit() throws OdetteFTPException {

		OdetteExchangeBuffer cdt = receive();

		if (cdt.getIdentifier() == CommandIdentifier.CDT) {
			resetCredits();

		} else {
			abnormalRelease(EndSessionReason.PROTOCOL_VIOLATION);
		}

	}

	public void closeFile(int recordCount, long unitCount)
			throws OdetteFTPException {

		Session context = getContext();

		VirtualFile file = context.getFile();

		/**
		 * End File command.
		 */

		if ((file.getFormat() == RecordFormat.UNSTRUCTURED)
				|| (file.getFormat() == RecordFormat.TEXTFILE)) {

			recordCount = 0;
		}

		CommandExchangeBuffer efid;
		efid = CommandExchangeBuffer.endFile(recordCount, unitCount);

		send(efid);

		/**
		 * Receiving the End File Answer command.
		 */
		CommandExchangeBuffer answer = (CommandExchangeBuffer) receive();
		CommandIdentifier ci = answer.getIdentifier();

		if ((ci == CommandIdentifier.EFPA) || (ci == CommandIdentifier.EFNA)) {

		} else {
			abnormalRelease(EndSessionReason.PROTOCOL_VIOLATION,
					"EFPA or EFNA commands is expected");
		}

		boolean turnDirection = false;

		if (ci == CommandIdentifier.EFPA) {
			turnDirection = answer.getParameter("EFPACD").equals("Y");

		} else if (ci == CommandIdentifier.EFNA) {
			String code = answer.getParameter("EFNAREAS");
			AnswerReason reason = (AnswerReason) AnswerReason.parse(code);
			throw new FileTransferException(reason, "End File Negative "
					+ "Answer: " + code);
		}

		if (turnDirection) {
			changeDirection();
		}

	}

	public void changeDirection() throws OdetteFTPException {

		Session context = getContext();
		PreferencesUtil prefs = getPreferences();

		if (prefs.getTransferMode() == TransferMode.SENDER_ONLY) {
			abnormalRelease(EndSessionReason.INCOMPATIBLE_MODE, "CD command "
					+ "not acceptable in the Sender Only transfer mode");
		}

		/**
		 * Sending Change Direction command.
		 */
		CommandExchangeBuffer cd = CommandExchangeBuffer.changeDirection();

		send(cd);

		context.changeState(ListenerState.class);
		context.state().listen();
	}

	public void endToEndResponse(String datasetName, Date fileDateTime,
			String destination, String originator, String userData,
			String reserved) throws OdetteFTPException {

		CommandExchangeBuffer eerp;

		/**
		 * Send EERP command.
		 */

		/* F_EERP_RQ */
		eerp = CommandExchangeBuffer.endToEndResponse(datasetName, "",
				fileDateTime, userData, destination, originator);

		send(eerp);

		/**
		 * Wait for receiving the Ready to Receive command.
		 */
		OdetteExchangeBuffer answer = receive();

		/* violation */
		if (answer.getIdentifier() != CommandIdentifier.RTR) {
			abnormalRelease(EndSessionReason.PROTOCOL_VIOLATION,
					"RTR command is expected");
		}
	}

	private void illegalState() throws OdetteFTPException {
		throw new IllegalStateException("Speaker state doesn't support this "
				+ "operation");
	}

	private void waitForReadyMessage() throws OdetteFTPException {

		// receive ssrm command
		OdetteExchangeBuffer ssrm = receive();

		// check if is corresponding identifier
		if (ssrm.getIdentifier() != CommandIdentifier.SSRM) {
			abnormalRelease(EndSessionReason.PROTOCOL_VIOLATION,
					"SSRM command expected");
		}
	}

	private void negotiateStartSession(String userCode, String password,
			TransferMode mode, boolean restart) throws OdetteFTPException {

		Session context = getContext();
		PreferencesUtil prefs = getPreferences();

		/*
		 * Retrieve and check session parameters.
		 */
		int sdeb = prefs.getBufferSize();
		boolean compression = prefs.isCompressionSupported();
		boolean specialLogic = prefs.hasSpecialLogic();
		int window = prefs.getWindowSize();
		String userData = prefs.getUserData();

		checkStartSession(context, sdeb, window, compression, mode, restart);

		/*
		 * Sending Start Session command.
		 */
		String reserved = "";
		CommandExchangeBuffer ssid = CommandExchangeBuffer.startSession(
				userCode, password, sdeb, mode, compression, restart,
				specialLogic, window, reserved, userData);

		send(ssid);

		/*
		 * Receiving Start Session command.
		 */
		CommandExchangeBuffer receivedCmd = (CommandExchangeBuffer) receive();

		if (receivedCmd.getIdentifier() == CommandIdentifier.SSID) {

			/* check values negotiated on session and do set up */
			sessionSetup(userCode, password, sdeb, mode, compression, restart,
					specialLogic, window, receivedCmd);

		} else {
			abnormalRelease(EndSessionReason.PROTOCOL_VIOLATION,
					"SSID commands were expected");
		}
	}

	private void sessionSetup(String userCode, String password, int sdeb,
			TransferMode mode, boolean compression, boolean restart,
			boolean specialLogic, int window, CommandExchangeBuffer ssid)
			throws OdetteFTPException {

		Session context = getContext();
		PreferencesUtil prefs = getPreferences();

		String ssidcode = ssid.getParameter("SSIDCODE");
		String ssidpswd = ssid.getParameter("SSIDPSWD");
		int ssidsdeb = Integer.parseInt(ssid.getParameter("SSIDSDEB"));
		TransferMode ssidsr = (TransferMode) TransferMode.parse(ssid
				.getParameter("SSIDSR"));
		boolean ssidcmpr = valueOfYesNo(ssid.getParameter("SSIDCMPR"));
		boolean ssidrest = valueOfYesNo(ssid.getParameter("SSIDREST"));
		boolean ssidspec = valueOfYesNo(ssid.getParameter("SSIDSPEC"));
		int ssidcred = Integer.parseInt(ssid.getParameter("SSIDCRED"));

		if ((!compression) && (ssidcmpr)) {
			abnormalRelease(EndSessionReason.INCOMPATIBLE_MODE,
					"peer used compression");
		}

		if ((!restart) && (ssidrest)) {
			abnormalRelease(EndSessionReason.INCOMPATIBLE_MODE,
					"peer used restart");
		}

		TransferMode ssidsrReversed = ssidsr.getReversed();
		if ((mode != TransferMode.BOTH) && (mode != ssidsrReversed)) {
			abnormalRelease(EndSessionReason.INCOMPATIBLE_MODE,
					"mode must be: " + mode);
		}

		if (specialLogic != ssidspec) {
			abnormalRelease(EndSessionReason.INCOMPATIBLE_MODE,
					"wrong special logic indicator specified by peer");
		}

		if ((ssidsdeb > sdeb) || bufferSizeViolation(ssidsdeb, context)) {
			abnormalRelease(EndSessionReason.INVALID_COMMAND_DATA, "peer "
					+ "used a invalid exchange buffer size: " + ssidsdeb);
		}

		if ((ssidcred > window) || windowSizeViolation(ssidcred, context)) {
			abnormalRelease(EndSessionReason.INVALID_COMMAND_DATA, "peer "
					+ "used a invalid window size (credits): " + ssidcred);
		}

		prefs.setCompressionSupport(ssidcmpr);
		prefs.setRestartSupport(ssidrest);
		prefs.setTransferMode(ssidsrReversed);
		prefs.setSpecialLogic(ssidspec);
		prefs.setBufferSize(ssidsdeb);
		prefs.setWindowSize(ssidcred);

		/* check returning credentials */
		requestAuthentication(ssidcode, ssidpswd, prefs.getCalledAddress());
	}

	protected void setCredits(int windowSize) {
		Session context = getContext();
		context.setProperty("odette.credits.speaker", String
				.valueOf(windowSize));
	}

	public int getCredits() {
		Session context = getContext();
		String credits = context.getProperty("odette.credits.speaker");

		return Integer.parseInt(credits);
	}

}
