/**
 *  OdetteJ - Open Source Odette FTP implementation in Java(TM).
 *  Copyright (C) 2005 Fossil E-Commerce, http://www.fossilec.com
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  ------------------------------------------------------------------------
 *  $Id: CommandExchangeBuffer.java,v 1.10 2005/09/12 19:17:27 rmarins Exp $
 *  ------------------------------------------------------------------------
 */

package org.fossilec.odettej.service;

import java.io.UnsupportedEncodingException;
import java.nio.ByteBuffer;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.Iterator;

import org.fossilec.odettej.OdetteFTPException;
import org.fossilec.odettej.StringUtil;
import org.fossilec.odettej.service.CommandFormat.Field;

/**
 * A Command Exchange Buffer contains a single command starting at the beginning
 * of the buffer. Commands and data are never mixed within an Exchange Buffer.
 * Each command has a fixed length and can not be compressed.
 */
public class CommandExchangeBuffer extends OdetteExchangeBuffer {

	/**
	 * Carriage return string used in some command. The specification determine
	 * it could be defined as 0x0a or 0x08.
	 */
	private static final String PROTOCOL_CARRIAGE_RETURN = "\n";

	/**
	 * String formatter pattern for converting a <code>Date</code> value to
	 * the Date stamp used on commands.
	 * 
	 * @see SimpleDateFormat#SimpleDateFormat(java.lang.String)
	 */
	private static final String DATE_STAMP_PATTERN = "yyMMdd";

	/**
	 * String formatter pattern for converting a <code>Date</code> value to
	 * the Time stamp used on commands.
	 * 
	 * @see SimpleDateFormat#SimpleDateFormat(java.lang.String)
	 */
	private static final String TIME_STAMP_PATTERN = "HHmmss";

	/**
	 * Default charset defined in Odette FTP protocol specification.
	 */
	private static final String DEFAULT_PROTOCOL_CHARSET = "ISO_646.IRV:1991";

	/**
	 * Initialized with the command format of the respective identifier.
	 */
	private final CommandFormat format;

	/**
	 * When set to true, the Command Exchange Buffer throw error if parsing a
	 * command with wrong format (missing the carriage return, for instance).
	 */
	private boolean checkFormatStrictly;

	protected CommandExchangeBuffer(ByteBuffer rawBuffer, boolean strictFormat)
			throws OdetteFTPException {

		super();

		checkFormatStrictly = strictFormat;
		parse(rawBuffer);

		format = CommandFormat.getFormat(getIdentifier());
	}

	/**
	 * Default constructor.
	 * 
	 * @param identifier
	 *            Command identifier.
	 * @throws OdetteFTPException
	 */
	private CommandExchangeBuffer(CommandIdentifier identifier)
			throws OdetteFTPException {

		super();
		int commandSize;

		format = CommandFormat.getFormat(identifier);
		commandSize = format.getLength();

		buffer = ByteBuffer.allocateDirect(commandSize);

		setLength(commandSize);
		setIdentifier(identifier);
	}

	/**
	 * Start Session Ready Message.
	 * 
	 * @return Corresponding <code>CommandExchangeBuffer</code>.
	 * @throws OdetteFTPException
	 */
	public static CommandExchangeBuffer readyMessage()
			throws OdetteFTPException {
		CommandExchangeBuffer ssrm = new CommandExchangeBuffer(
				CommandIdentifier.SSRM);

		ssrm.setParamater("SSRMMSG", "ODETTE FTP READY ");
		ssrm.setParamater("SSRMCR", PROTOCOL_CARRIAGE_RETURN);

		return ssrm;
	}

	/**
	 * Create the Start Session command with given parameters. It belongs to the
	 * Start Session Phase, and is performed in both direction to negotiate
	 * capabilities and session wide parameters between locations.
	 * 
	 * @param code
	 *            Initiator's Identification Code which uniquely identifies the
	 *            Initiator (sender) participating in the Odette FTP session.
	 * @param pswd
	 *            Key to authenticate the sender. Assigned by bilateral
	 *            agreement.
	 * @param sdeb
	 *            The length, in octets, of the largest Exchange Buffer that can
	 *            be accepted by the location.
	 * @param sendReceive
	 *            Sender / Receiver capabilities:
	 * @param compression
	 *            Compression indicator. <code>true</code> if the location can
	 *            handle compressed data. Otherwise it should be
	 *            <code>false</code>.
	 * @param restart
	 *            Restart indication informing whether the location can handle
	 *            the restart of a partially transmitted file.
	 * @param specialLogic
	 *            Special logic indication.
	 * @param credit
	 *            Credit.
	 * @param reserved
	 *            Reserved for future use.
	 * @param userData
	 *            User Data.
	 * @return The Start Session command with the corresponding values.
	 * @throws OdetteFTPException
	 */
	public static CommandExchangeBuffer startSession(String code, String pswd,
			int sdeb, TransferMode mode, boolean compression, boolean restart,
			boolean specialLogic, int credit, String reserved, String userData)
			throws OdetteFTPException {

		CommandExchangeBuffer ssid = new CommandExchangeBuffer(
				CommandIdentifier.SSID);

		ssid.setParamater("SSIDLEV", "1");
		ssid.setParamater("SSIDCODE", code);
		ssid.setParamater("SSIDPSWD", pswd);
		ssid.setParamater("SSIDSDEB", String.valueOf(sdeb));
		ssid.setParamater("SSIDSR", String.valueOf(mode.getCode()));
		ssid.setParamater("SSIDCMPR", yesNo(compression));
		ssid.setParamater("SSIDREST", yesNo(restart));
		ssid.setParamater("SSIDSPEC", yesNo(specialLogic));
		ssid.setParamater("SSIDCRED", String.valueOf(credit));
		ssid.setParamater("SSIDRSV1", reserved);
		ssid.setParamater("SSIDUSER", userData);
		ssid.setParamater("SSIDCR", PROTOCOL_CARRIAGE_RETURN);

		return ssid;
	}

	/**
	 * Create the Start File command with given parameters.<br>
	 * 
	 * The Start File command includes a count allowing the restart of an
	 * interrupted transmission to be negotiated. If restart facilities are not
	 * available the restart count must be set to zero. The sender will start
	 * with the lowest record count + 1.
	 * 
	 * @param datasetName
	 *            Dataset name of the Virtual File being transferred assigned by
	 *            bilateral agreement.
	 * @param reserved
	 *            This field is reserved for future use.
	 * @param dateTime
	 *            Specific Date and Time assigned by the Virtual File's
	 *            Originator indicating when the file was made avaiable for
	 *            transmission.
	 * @param userData
	 *            May be used by the Odette FTP in any way. If unused it should
	 *            be initialized to spaces. It is expected that a bilateral
	 *            agreement exists as to the meaning of the data.
	 * @param destination
	 *            The Identification Code for the final recipient of the Virtual
	 *            File. This is the location that will look into the Virtual
	 *            File content and perform mapping functions. It is also the
	 *            location that creates the End to End Response (EERP) command
	 *            for the received file.
	 * @param originator
	 *            The Identification Code from the Originator of the Virtual
	 *            File. It is the location that created (mapped) the data for
	 *            transmission.
	 * @param recordFormat
	 *            Virtual File format (Fixed, Variable, Unustructured, Text
	 *            File). Used to calculate the restart position.
	 * @param maxRecordSize
	 *            Length in octets of the longest logical record which may be
	 *            transferred to a location. Only user data is included. If File
	 *            format is 'T' or 'U' then this attribute must be set to
	 *            '00000'
	 * @param fileSize
	 *            File Size, 1K (1024 octets) blocks.
	 * @param restartOffset
	 *            Restart position.
	 * @return The Start File command with the corresponding values.
	 * @throws OdetteFTPException
	 */
	public static CommandExchangeBuffer startFile(String datasetName,
			String reserved, Date dateTime, String userData,
			String destination, String originator, RecordFormat recordFormat,
			int maxRecordSize, int fileSize, int restartOffset)
			throws OdetteFTPException {

		CommandExchangeBuffer sfid = new CommandExchangeBuffer(
				CommandIdentifier.SFID);

		sfid.setParamater("SFIDDSN", datasetName);
		sfid.setParamater("SFIDRSV1", reserved);
		sfid.setParamater("SFIDDATE", StringUtil.formatDate(DATE_STAMP_PATTERN,
				dateTime));
		sfid.setParamater("SFIDTIME", StringUtil.formatDate(TIME_STAMP_PATTERN,
				dateTime));
		sfid.setParamater("SFIDUSER", userData);
		sfid.setParamater("SFIDDEST", destination);
		sfid.setParamater("SFIDORIG", originator);
		sfid.setParamater("SFIDFMT", recordFormat.toString());
		sfid.setParamater("SFIDLRECL", String.valueOf(maxRecordSize));
		sfid.setParamater("SFIDFSIZ", String.valueOf(fileSize));
		sfid.setParamater("SFIDREST", String.valueOf(restartOffset));

		return sfid;
	}

	/**
	 * Create the Start File Positive Answer command. The only parameter
	 * indicate which position the Listener agree to restart the receive of a
	 * previous Virtual File.<br>
	 * 
	 * @param answerCount
	 *            <code>int</code> lower or equal to restart count specified
	 *            by the Speaker in the Start File (SFID) command. If restart
	 *            facilities are not avaiable, a count of zero must be
	 *            specified.
	 * @return The Start File Positive Answer command with the corresponding
	 *         values.
	 * @throws OdetteFTPException
	 */
	public static CommandExchangeBuffer startFilePositiveAnswer(int answerCount)
			throws OdetteFTPException {

		CommandExchangeBuffer sfpa = new CommandExchangeBuffer(
				CommandIdentifier.SFPA);

		sfpa.setParamater("SFPAACNT", String.valueOf(answerCount));

		return sfpa;
	}

	/**
	 * Create the Start File Negative Answer command containing the the reason
	 * why transmission can not proced.<br>
	 * 
	 * This <code>retry</code> parameter is used to advise the Speaker if it
	 * should retry at a latter point in time due to a temporary condition at
	 * the Listener site, such as a lack of storage space. It should be used in
	 * conjunction with the Answer Reason code.
	 * 
	 * @param reason
	 *            Answer Reason.
	 * @param retry
	 *            <code>true</code> if the transmission may be retried latter,
	 *            or <code>false</code> to don't retry again.
	 * @return The Start File Negative Answer command with the corresponding
	 *         values.
	 * @throws OdetteFTPException
	 */
	public static CommandExchangeBuffer startFileNegativeAnswer(
			AnswerReason reason, boolean retry) throws OdetteFTPException {

		CommandExchangeBuffer sfna = new CommandExchangeBuffer(
				CommandIdentifier.SFNA);

		sfna.setParamater("SFNAREAS", reason.getCode());
		sfna.setParamater("SFNARRTR", yesNo(retry));

		return sfna;
	}

	/**
	 * Create the Set Credit command which is used to avoid congestion at the
	 * protocol level a flow control.
	 * 
	 * @param reserved
	 *            This field is reserved for future use.
	 * @return The Set Credit command with the corresponding values.
	 * @throws OdetteFTPException
	 */
	public static CommandExchangeBuffer setCredit(String reserved)
			throws OdetteFTPException {

		CommandExchangeBuffer cdt = new CommandExchangeBuffer(
				CommandIdentifier.CDT);

		cdt.setParamater("CDTRSV1", reserved);

		return cdt;
	}

	/**
	 * Create the End File command indicating the count of records and overall
	 * octets from the transmitted Virtual File.<br>
	 * 
	 * The count will express the real size of the file (before compression,
	 * header not included). The total count is always used, even during restart
	 * processing.
	 * 
	 * @param recordCount
	 *            For FIXED or VARIABLE record formats, the exact record count.
	 *            For UNSTRUCTURED or TEXTFILE, must be zero.
	 * @param unitCount
	 *            Exact number of units (octets) transmitted.
	 * @return
	 * @throws OdetteFTPException
	 */
	public static CommandExchangeBuffer endFile(int recordCount, long unitCount)
			throws OdetteFTPException {

		CommandExchangeBuffer efid = new CommandExchangeBuffer(
				CommandIdentifier.EFID);

		efid.setParamater("EFIDRCNT", String.valueOf(recordCount));
		efid.setParamater("EFIDUCNT", String.valueOf(unitCount));

		return efid;
	}

	/**
	 * Create the End File Positive Answer command indicating whether the
	 * Listener is requesting or not a Change Direction (CD) command from
	 * Speaker or not.
	 * 
	 * @param changeDirection
	 *            Change Direction Indicator.
	 * @return The End File Positive Answer command with the corresponding
	 *         values.
	 * @throws OdetteFTPException
	 */
	public static CommandExchangeBuffer endFilePositiveAnswer(
			boolean changeDirection) throws OdetteFTPException {

		CommandExchangeBuffer efpa = new CommandExchangeBuffer(
				CommandIdentifier.EFPA);

		efpa.setParamater("EFPACD", yesNo(changeDirection));

		return efpa;
	}

	/**
	 * Create the End File Negative Answer command providing the reason for
	 * stopping transfer.
	 * 
	 * @param reason
	 *            Reason why transmission can not proceed.
	 * @return The End File Negative Answer command with the corresponding
	 *         values.
	 * @throws OdetteFTPException
	 */
	public static CommandExchangeBuffer endFileNegativeAnswer(
			AnswerReason reason) throws OdetteFTPException {

		CommandExchangeBuffer efna = new CommandExchangeBuffer(
				CommandIdentifier.EFNA);

		efna.setParamater("EFNAREAS", reason.getCode());

		return efna;
	}

	/**
	 * Create the End Session command indicating the reason code for terminating
	 * the session.
	 * 
	 * @param reason
	 *            End Session Reason.
	 * @return The End Session command with the corresponding values.
	 * @throws OdetteFTPException
	 */
	public static CommandExchangeBuffer endSession(EndSessionReason reason)
			throws OdetteFTPException {

		CommandExchangeBuffer esid = new CommandExchangeBuffer(
				CommandIdentifier.ESID);

		esid.setParamater("ESIDREAS", reason.getCode());
		esid.setParamater("ESIDCR", PROTOCOL_CARRIAGE_RETURN);

		return esid;
	}

	/**
	 * Create the Change Direction command.
	 * 
	 * @return The Change Direction command with the corresponding values.
	 * @throws OdetteFTPException
	 */
	public static CommandExchangeBuffer changeDirection()
			throws OdetteFTPException {
		CommandExchangeBuffer cd = new CommandExchangeBuffer(
				CommandIdentifier.CD);
		return cd;
	}

	/**
	 * Create the End to End Response command with given parameters.
	 * 
	 * @param dataSetName
	 *            Dataset name of the Virtual File being transfered.
	 * @param reserved
	 *            This field is reserved for future use.
	 * @param dateTime
	 *            Virtual File date and time indicating when the file was made
	 *            available for transmission.
	 * @param userData
	 *            May be used by the Odette FTP in any way.
	 * @param destination
	 *            Identification code from the Originator of the Virtual File,
	 *            which created (mapped) the data for transmission.
	 * @param originator
	 *            Identification code of the Final Recipient of the Virtual
	 *            File. This is the location that creates the EERP for the
	 *            received file.
	 * @return The End to End Response command with the corresponding values.
	 * @throws OdetteFTPException
	 */
	public static CommandExchangeBuffer endToEndResponse(String dataSetName,
			String reserved, Date dateTime, String userData,
			String destination, String originator) throws OdetteFTPException {

		CommandExchangeBuffer eerp = new CommandExchangeBuffer(
				CommandIdentifier.EERP);

		eerp.setParamater("EERPDSN", dataSetName);
		eerp.setParamater("EERPRSV1", reserved);
		eerp.setParamater("EERPDATE", StringUtil.formatDate(DATE_STAMP_PATTERN,
				dateTime));
		eerp.setParamater("EERPTIME", StringUtil.formatDate(TIME_STAMP_PATTERN,
				dateTime));
		eerp.setParamater("EERPUSER", userData);
		eerp.setParamater("EERPDEST", destination);
		eerp.setParamater("EERPORIG", originator);

		return eerp;
	}

	/**
	 * Create the Ready to Receive command.
	 * 
	 * @return The Ready to Receive command with the corresponding values.
	 * @throws OdetteFTPException
	 */
	public static CommandExchangeBuffer readyToReceive()
			throws OdetteFTPException {
		CommandExchangeBuffer rtr = new CommandExchangeBuffer(
				CommandIdentifier.RTR);
		return rtr;
	}

	/**
	 * Convenient method for returning Yes or No (Y/N) to a given
	 * <code>boolean</code> value. It almost used to fill YesNo fields on
	 * commands.
	 * 
	 * @param set
	 *            The boolean value. Y = <code>true</code>, N =
	 *            <code>false</code>.
	 * @return <code>String</code> "Y" or "N" representing given value.
	 */
	private static String yesNo(boolean set) {
		return (set ? "Y" : "N");
	}

	/**
	 * Return the Charset Name used for conversion between Strings and its Bytes
	 * representation on buffer.
	 * 
	 * @return Charset Name.
	 */
	private String getCharsetName() {
		return DEFAULT_PROTOCOL_CHARSET;
	}

	/**
	 * Return a Command Exchange Buffer parameter value for the specified field.
	 * 
	 * @param name
	 *            Field key name.
	 * @return Corresponding parameter value for a given field.
	 */
	public String getParameter(String name) {
		String value = null;

		if (format.containsField(name)) {
			Field field = format.getField(name);
			ByteBuffer buffer = getBuffer();

			int startPosition = field.getPosision();
			int paramSize = field.getSize();
			byte octets[] = new byte[paramSize];

			buffer.position(startPosition);
			buffer.get(octets, 0, paramSize);

			value = parseOctects(octets);

			if (field.getType() == CommandFormat.ALPHANUMERIC_TYPE
					&& field.getFormat() == CommandFormat.VARIABLE_FORMAT) {
				value = value.trim();
			}
		}

		return value;
	}

	/**
	 * @return Collection of strings representing parameters.
	 */
	public Collection getParameters() {
		Collection parms = new ArrayList();

		Iterator it = format.getFields().iterator();
		while (it.hasNext()) {
			Field field = (Field) it.next();
			parms.add(field.getName());
		}

		return parms;
	}

	/**
	 * Set a parameter in command stream with the given value.
	 * 
	 * @param name
	 *            parameter name (as defined in protocol specification).
	 * @param value
	 *            <code>String</code> representing the given value as expected
	 *            by protocol specification.
	 * @throws OdetteFTPException
	 */
	protected void setParamater(String name, String value)
			throws OdetteFTPException {

		Field field = format.getField(name);

		checkParameterFormat(field, name, value);
		String formattedValue = formatParameter(field, value);

		byte[] octets = parseString(formattedValue);

		BufferUtil.putAtPosition(super.buffer, octets, field.getPosision());
	}

	/**
	 * Check if a given parameter follows its length and type (in case of
	 * numeric fields).
	 * 
	 * @param field
	 *            object representing Command's field to query checking format.
	 * @param name
	 *            parameter name (as defined in protocol specification).
	 * @param value
	 *            <code>String</code> representing the given value as expected
	 *            by protocol specification.
	 * @throws OdetteFTPException
	 * @throws InvalidCommandDataException
	 *             When the specified value doesn't correspond to the required
	 *             field format.
	 */
	private static void checkParameterFormat(Field field, String name,
			String value) throws OdetteFTPException {

		if (value == null) {
			return;
		}

		int length = value.length();
		if (length > field.getSize()) {
			throw new InvalidCommandDataException("Invalid parameter length: "
					+ name + " length is " + length);
		}

		char type = field.getType();
		if (type == CommandFormat.NUMERIC_TYPE) {
			if (!isNumeric(value)) {
				throw new InvalidCommandDataException(
						"Parameter is not numeric: " + name + " = " + value);
			}
		}
	}

	/**
	 * Given a <code>String</code> value, return <code>true</code> if it
	 * contains only numbers or <code>false</code> if it's a alphanumeric
	 * String.
	 * 
	 * @param value
	 *            <code>String</code> value.
	 * @return <code>boolean</code> Given value is numeric.
	 */
	private static boolean isNumeric(String value) {
		try {
			Long.parseLong(value);
			return true;
		} catch (Exception e) {
			return false;
		}
	}

	/**
	 * Return the formatted value to be set on a command field, with respect to
	 * it's format.
	 * 
	 * @param field
	 *            Command field format definition.
	 * @param value
	 *            Value to be formatted.
	 * @return Formatted value.
	 * @throws OdetteFTPException
	 */
	private static String formatParameter(Field field, String value)
			throws OdetteFTPException {

		String result = null;

		char type = field.getType();
		int length = field.getSize();

		if (type == CommandFormat.ALPHANUMERIC_TYPE) {
			result = StringUtil.padd(value, length, false, ' ');
		} else if (type == CommandFormat.NUMERIC_TYPE) {
			result = StringUtil.padd(value, length, true, '0');
		}

		return result;
	}

	/**
	 * Transform a given <code>String</code> value to the corresponding octets
	 * array. A charset conversion is performed using the
	 * <code>getCharsetName</code> method.
	 * 
	 * @param value
	 * @return
	 */
	private byte[] parseString(String value) {
		try {
			return value.getBytes(getCharsetName());
		} catch (UnsupportedEncodingException e) {
			return null;
		}
	}

	/**
	 * Transform a given array of octets to <code>String</code> using some
	 * charset conversion scheme. Charset is obtained from the
	 * <code>getCharsetName</code> method.
	 * 
	 * @param octets
	 * @return
	 */
	private String parseOctects(byte[] octets) {
		try {
			return new String(octets, getCharsetName());
		} catch (UnsupportedEncodingException e) {
			return null;
		}
	}

	/**
	 * Method used to set configure this object with the value from parsing the
	 * given raw buffer. Raw buffer usually comes directly from transportation
	 * stream.
	 * 
	 * @param rawBuffer
	 *            raw command exchange buffer to parse.
	 * @throws BufferCorruptedException
	 *             When buffer has a invalid remaing size, or there's a wrong
	 *             command length (and the strictly check format option is set).
	 * @throws OdetteFTPException
	 */
	protected void parse(ByteBuffer rawBuffer) throws OdetteFTPException {

		int remaining;

		if (rawBuffer == null) {
			throw new IllegalArgumentException("Exchange Buffer is null");
		}

		remaining = rawBuffer.remaining();
		if (remaining == 0) {
			throw new BufferCorruptedException(
					"invalid remaining buffer size: " + remaining);
		}

		setBuffer(rawBuffer.slice());

		CommandIdentifier cmd = getIdentifier();
		int commandSize = CommandFormat.getFormat(cmd).getLength();
		if ((remaining != commandSize) && (checkFormatStrictly)) {
			throw new BufferCorruptedException("Exchange Buffer size error");
		}

		setLength(remaining);
	}
}
