/**
 *  OdetteJ - Open Source Odette FTP implementation in Java(TM).
 *  Copyright (C) 2005 Fossil E-Commerce, http://www.fossilec.com
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  ------------------------------------------------------------------------
 *  $Id: ClassloaderTransportFactory.java,v 1.1 2005/08/15 16:44:36 rmarins Exp $
 *  ------------------------------------------------------------------------
 */

package org.fossilec.odettej.transport;

import java.lang.reflect.Constructor;

/**
 * Object Factory used to instantiate <code>Transport</code> implementations
 * using a full qualified class name. The target implementation class MUST
 * provide a default constructor method.
 * <p>
 * The ClassloaderTransportFactory takes advantages in use reflections for
 * object instantiation. A common Classloader is used from the current running
 * Thread.
 * 
 * @author Rafael Marins
 * @see org.fossilec.odettej.transport.TransportFactory
 */
public class ClassloaderTransportFactory extends TransportFactory {

	/**
	 * Use default constructor with no parameters for <code>Transport</code>
	 * object instatiation.
	 */
	private static final Class[] TRANSPORT_DEFAULT_PARAM = new Class[] {};

	private String className;

	/**
	 * Factory constructor with the target full qualified class name in the
	 * parameter, used to create <code>Transport</code> implementations.
	 * 
	 * @param className
	 *            fully qualified class name implementing the
	 */
	public ClassloaderTransportFactory(String className) {
		super();
		this.className = className;
	}

	/**
	 * Create a <code>Transport</code> class instance using the path provided
	 * in the constructor. Return <code>null</code> in any of the following
	 * cases:
	 * <ul>
	 * <li>Class not found.
	 * <li>Target class not right implemented (maybe abstract).
	 * <li>Default constructor doesn't exist or without required method
	 * visibility.
	 * <li>
	 * </ul>
	 * 
	 * @see org.fossilec.odettej.transport.TransportFactory#createTransport()
	 * @return An object implementing the <code>Transport</code> interface.
	 */
	public Transport createTransport() {

		Transport transport = null;
		ClassLoader cl = getClassLoader();

		try {
			Class clazz = cl.loadClass(className);
			Constructor ctr = clazz.getConstructor(TRANSPORT_DEFAULT_PARAM);
			transport = (Transport) ctr.newInstance(new Object[] {});
		} catch (Exception e) {
			// nothing to do
		}

		return transport;
	}

	/**
	 * Return the common Classloader from the current running Thread.
	 * 
	 * @return <code>ClassLoader</code> object.
	 */
	private static ClassLoader getClassLoader() {
		ClassLoader cl = Thread.currentThread().getContextClassLoader();
		if (cl == null) {
			cl = Transport.class.getClassLoader();
		}
		return cl;
	}

}
