/**
 *  OdetteJ - Open Source Odette FTP implementation in Java(TM).
 *  Copyright (C) 2005 Fossil E-Commerce, http://www.fossilec.com
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  ------------------------------------------------------------------------
 *  $Id: Transport.java,v 1.10 2005/09/10 23:59:27 rmarins Exp $
 *  ------------------------------------------------------------------------
 */

package org.fossilec.odettej.transport;

import java.nio.ByteBuffer;

import org.fossilec.odettej.OdetteFTPException;

/**
 * This interface provides a protocol independent network transport layer
 * required by Odette FTP peer entities to with each via the underlying OSI
 * model Network Service provider.
 * 
 * @author Rafael Marins
 */
public interface Transport {

	/**
	 * Setup a network connection with another Odette FTP peer entity.
	 * 
	 * @param address
	 *            A protocol specific <code>NetworkAddress</code>
	 *            implementation with the network location of the target entity.
	 * @return <code>true</code> when connection is established with success
	 * @throws ConnectionException
	 *             If the transport instance is already connected.
	 * @throws AddressException
	 *             For unresolved or unsupported address types
	 * @throws OdetteFTPException
	 *             Thrown by unexpected exception
	 */
	public abstract boolean connect(NetworkAddress address)
			throws OdetteFTPException;

	/**
	 * Setup a connection with another Odette FTP peer entity.
	 * 
	 * @param address
	 *            A String representation of the network location of the target
	 *            entity of the underlying transport protocol. The value is
	 *            parsed by the corresponding <code>NetworkAddress</code>
	 *            implementation of that protocol.
	 * @return <code>true</code> when connection is established with success
	 * @throws ConnectionException
	 *             If the transport instance is already connected.
	 * @throws AddressException
	 *             For unresolved or unsupported address types.
	 * @throws OdetteFTPException
	 *             Thrown for unexpected failures
	 */
	public abstract boolean connect(String address) throws OdetteFTPException;

	/**
	 * Terminate the ongoing connection. Once the connection is ended, it's not
	 * avaiable for further communication activity.
	 * 
	 * @throws OdetteFTPException
	 */
	public abstract void disconnect() throws OdetteFTPException;

	/**
	 * Returns local address used to fulfill Odette FTP session variables. On
	 * the <i>Initiator</i> side, it compose the "<code>odette.calling-address
	 * </code>"
	 * property. The <i>Responder</i> Odette FTP entity use this method to fill
	 * the "<code>odette.called-address</code>" property.
	 * 
	 * @return String representing the local address of the underlying network
	 *         protocol, as expected by session variables.
	 * @throws OdetteFTPException
	 */
	public abstract String getLocalAddress() throws OdetteFTPException;

	/**
	 * Returns remote address used to fulfill Odette FTP session variables. On
	 * the <i>Initiator</i> side, it's used in the "<code>odette.called-address
	 * </code>".
	 * The <i>Responder</i> Odette FTP entity use this method to fill the "<code>odette.calling-address</code>"
	 * property.
	 * 
	 * @return String representing the remote address of the underlying network
	 *         protocol, as expected by session variables.
	 * @throws OdetteFTPException
	 */
	public abstract String getRemoteAddress() throws OdetteFTPException;

	/**
	 * Tells whether the underlying network protocol is connected.
	 * 
	 * @return <code>true</code> if this transport implementation is
	 *         connected.
	 * @throws OdetteFTPException
	 */
	public abstract boolean isConnected() throws OdetteFTPException;

	/**
	 * Receive a sequence of bytes from the underlying transport implementation
	 * of network protocol.
	 * <p>
	 * An attempt is made to receive up to <i>r</i> bytes from the channel,
	 * where <i>r</i> is the number of bytes remaining in the buffer, that is,
	 * <code>dst.remaining()</code>, at the moment this method is invoked.
	 * 
	 * @param dst
	 *            The buffer into which bytes are to be transferred
	 * @return The number of bytes received, possibly zero, or <code>-1</code>
	 *         if the channel has reached end-of-stream
	 * @throws NotConnectedException
	 *             Transport not yet connected
	 * @throws OdetteFTPException
	 *             Thrown by unexpected exception
	 */
	public abstract int receive(ByteBuffer dst) throws OdetteFTPException;

	/**
	 * Send a sequence of bytes through this transport implementation, of the
	 * underlying network protocol, from the given buffer.
	 * <p>
	 * An attempt is made to send up to <i>r</i> bytes to the channel, where
	 * <i>r</i> is the number of bytes remaining in the buffer, that is,
	 * <code>dst.remaining()</code>, at the moment this method is invoked.
	 * 
	 * @param src
	 *            The buffer from which bytes are to be retrieved
	 * @return The number of bytes sent, possibly zero
	 * @throws NotConnectedException
	 *             Transport not yet connected
	 * @throws OdetteFTPException
	 *             Thrown by unexpected exception
	 */
	public abstract int send(ByteBuffer src) throws OdetteFTPException;

	/**
	 * Sent a sequence of bytes through this transport implementation, of the
	 * underlying network protocol, from a subsequence of the given buffers.
	 * 
	 * @param srcs
	 *            The sequence of buffers from which bytes are to be retrieved
	 * @return The number of overall bytes sent, possibly zero
	 * @throws OdetteFTPException
	 */
	public abstract int send(ByteBuffer[] srcs) throws OdetteFTPException;

}