/**
 *  OdetteJ - Open Source Odette FTP implementation in Java(TM).
 *  Copyright (C) 2005 Fossil E-Commerce, http://www.fossilec.com
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  ------------------------------------------------------------------------
 *  $Id: MoreDataBitTransport.java,v 1.1 2005/09/10 23:59:55 rmarins Exp $
 *  ------------------------------------------------------------------------
 */

package org.fossilec.odettej.transport.mbgw;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.channels.SocketChannel;

import org.fossilec.odettej.OdetteFTPException;
import org.fossilec.odettej.service.OdetteExchangeBuffer;
import org.fossilec.odettej.transport.tcp.TCPTransport;

/**
 * This is a transport implementation of the More Data Bit protocol for
 * tunneling TCP/IP packages over X.25. It control the streaming sequence of
 * bytes over TCP/IP transmission using a header flag. In this implementation
 * it's done by overriding <code>send()</code> and <code>receive()</code>
 * methods from the <code>TCPTransport</code> class.
 * <p>
 * The More Data Bit protocol is required for using OdetteJ with TCP/IP - X.25
 * gateways. It have been specific used with <i>Access</i> equipments by
 * Microtronix (http://www.microtronix.com/).
 * 
 * @author Rafael Marins
 * @see org.fossilec.odettej.transport.Transport
 */
public class MoreDataBitTransport extends TCPTransport {

	private int receivingCount = 0;

	private ByteBuffer mbitHeader;

	/**
	 * Default constructor.
	 * 
	 * @see org.fossilec.odettej.transport.tcp.TCPTransport#TCPTransport()
	 */
	public MoreDataBitTransport() {
		super();
		initMbitHeader();
	}

	/**
	 * Creates a new class instance providing an existing SocketChannel.
	 * 
	 * @see org.fossilec.odettej.transport.tcp.TCPTransport#TCPTransport(SocketChannel)
	 */
	public MoreDataBitTransport(SocketChannel channel)
			throws OdetteFTPException {
		super(channel);
		initMbitHeader();
	}

	/**
	 * Allocate direct buffer in memory for handling More Data Bit headers.
	 */
	private void initMbitHeader() {
		mbitHeader = ByteBuffer.allocateDirect(2);
		mbitHeader.order(ByteOrder.BIG_ENDIAN);
	}

	/**
	 * Control the receiving More Data Bit packet headers using parent's
	 * <code>receive()</code> method. More information on this method function
	 * see <code>TCPTransport.receive(ByteBuffer)</code> for details.
	 * 
	 * @see org.fossilec.odettej.transport.tcp.TCPTransport#receive(ByteBuffer)
	 */
	public synchronized int receive(ByteBuffer dest) throws OdetteFTPException {

		if (receivingCount == 0) {

			mbitHeader.rewind();
			mbitHeader.clear();
			super.receive(mbitHeader);

			mbitHeader.rewind();
			receivingCount = mbitHeader.getShort();
		}

		int amount = 0;

		while ((receivingCount != 0) && (dest.hasRemaining())) {
			int count = super.receive(dest);
			receivingCount -= count;
			amount += count;
		}

		return amount;
	}

	/**
	 * Send a buffer using the More Data Bit control header. More information on
	 * this method function see <code>TCPTransport.send(ByteBuffer)</code>.
	 * 
	 * @see org.fossilec.odettej.transport.tcp.TCPTransport#send(ByteBuffer)
	 */
	public synchronized int send(ByteBuffer buffer) throws OdetteFTPException {

		OdetteExchangeBuffer oeb = OdetteExchangeBuffer.getInstance(buffer,
				false);
		mbitHeader.clear();
		mbitHeader.putShort((short) oeb.getLength());

		mbitHeader.rewind();
		super.send(mbitHeader);
		int count = super.send(oeb.getBuffer());

		return count;
	}

}
