/**
 *  OdetteJ - Open Source Odette FTP implementation in Java(TM).
 *  Copyright (C) 2005 Fossil E-Commerce, http://www.fossilec.com
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  ------------------------------------------------------------------------
 *  $Id: TCPTransport.java,v 1.13 2005/09/12 12:04:12 rmarins Exp $
 *  ------------------------------------------------------------------------
 */

package org.fossilec.odettej.transport.tcp;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.AlreadyConnectedException;
import java.nio.channels.IllegalBlockingModeException;
import java.nio.channels.NotYetConnectedException;
import java.nio.channels.SocketChannel;
import java.nio.channels.UnresolvedAddressException;
import java.nio.channels.UnsupportedAddressTypeException;

import org.fossilec.odettej.OdetteFTPException;
import org.fossilec.odettej.transport.AddressException;
import org.fossilec.odettej.transport.ConnectionException;
import org.fossilec.odettej.transport.NetworkAddress;
import org.fossilec.odettej.transport.NotConnectedException;
import org.fossilec.odettej.transport.Transport;

/**
 * Provide OdetteJ with a TCP/IP transportation layer based on Odette FTP,
 * implementing the <code>Transport</code> interface.
 * <p>
 * Used to create new transportion layer or encapsulating a existing opened
 * socket channel to provide its functionalities.
 * 
 * @author Rafael Marins
 * @see org.fossilec.odettej.transport.Transport
 */
public class TCPTransport implements Transport {

	/**
	 * Default TCP port used in Odette FTP if not specified when connecting.
	 */
	private static final int DEFAULT_PORT = 3305;

	/**
	 * Use a channel, New Java I/O API, to handle TCP socket stream.
	 */
	private SocketChannel socketChannel;

	/**
	 * Default constructor method. Create a new instance of SocketChannel for
	 * its use.
	 * 
	 * @throws OdetteFTPException
	 */
	public TCPTransport() {
		super();
	}

	/**
	 * Creates a new class instance providing an existing SocketChannel. This
	 * channel is opened in connection with another Odette FTP entity. It's
	 * useful when using OdetteJ for server mode operations.
	 * 
	 * @param channel
	 *            opened socket channel.
	 * @throws OdetteFTPException
	 */
	public TCPTransport(SocketChannel channel) throws OdetteFTPException {
		super();

		checkOpenedChannel(channel);
		this.socketChannel = channel;
	}

	/**
	 * Method used to verify required states in the channel param in class
	 * constructor.
	 * 
	 * @param channel
	 *            <code>SocketChannel</code> object parameter from class
	 *            constructor.
	 * @throws OdetteFTPException
	 */
	private static void checkOpenedChannel(SocketChannel channel)
			throws OdetteFTPException {

		if (channel == null) {
			throw new IllegalArgumentException("channel is null");
		} else if (!channel.isOpen()) {
			throw new IllegalArgumentException("channel is not opened");
		} else if (channel.isBlocking()) {
			throw new IllegalBlockingModeException();
		}
	}

	/**
	 * Gets channel used in streaming with TCP connection. Create a new
	 * <code>SocketChannel</code> if not set.
	 * 
	 * @return a channel.
	 * @throws OdetteFTPException
	 */
	protected SocketChannel getChannel() throws OdetteFTPException {
		if (socketChannel == null) {
			try {
				socketChannel = SocketChannel.open();
			} catch (IOException e) {
				throw new OdetteFTPException("Error opening channel", e);
			}
		}
		return socketChannel;
	}

	/**
	 * Setup a network connection with another Odette FTP entity.
	 * 
	 * @param address
	 *            <code>InternetAddress</code> type address providing the
	 *            network location to connect.
	 * @return <code>true</code> when connection is established with success,
	 *         otherwise return <code>false</code>.
	 * @throws ConnectionException
	 *             transport already connected
	 * @throws AddressException
	 *             unresolved address or unsupported address type
	 * @throws OdetteFTPException
	 *             unexpected error
	 * @see org.fossilec.odettej.transport.Transport#connect(NetworkAddress)
	 */
	public boolean connect(NetworkAddress address) throws OdetteFTPException {
		if (!(address instanceof InternetAddress)) {
			throw new AddressException("Illegal address type" + address);
		}
		InternetAddress addr = (InternetAddress) address;
		return protocolConnect(addr.getHost(), addr.getPort());
	}

	/**
	 * Setup a network connection with another Odette FTP entity.
	 * 
	 * @param address
	 *            A String representation of the TCP/IP network location of the
	 *            target Odette FTP peer. The value is parsed using
	 *            <code>InternetAddress</code> object.
	 * @return <code>true</code> when connection is established with success,
	 *         otherwise return <code>false</code>.
	 * @throws ConnectionException
	 *             transport already connected
	 * @throws AddressException
	 *             unresolved address or unsupported address type
	 * @throws OdetteFTPException
	 *             unexpected error
	 * @see org.fossilec.odettej.transport.Transport#connect(String)
	 */
	public boolean connect(String address) throws OdetteFTPException {
		InternetAddress addr = new InternetAddress(address);
		return protocolConnect(addr.getHost(), addr.getPort());
	}

	/**
	 * Delegation helper method used by both connect(...) procedures to
	 * establish the TCP connection with another Odette FTP peer. Is mandatory
	 * the use of host and port, even if above interface methods may not require
	 * then. In case it's not explicit in parameter, must use default TCP port
	 * for this service (odette-ftp).
	 * 
	 * @param host
	 *            Hostname or IP address of the Odette FTP entity to connect.
	 * @param port
	 *            Remote Odette FTP entity listening TCP port to connect.
	 * @return <code>true</code> if successfully connected, otherwise return
	 *         <code>false</code>.
	 * @throws ConnectionException
	 *             transport already connected
	 * @throws AddressException
	 *             unresolved address or unsupported address type
	 * @throws OdetteFTPException
	 *             unexpected error
	 */
	private synchronized boolean protocolConnect(String host, int port)
			throws OdetteFTPException {

		boolean connected = false;
		SocketChannel channel = getChannel();

		if (port <= 0) {
			port = DEFAULT_PORT;
		}

		try {
			connected = channel.connect(new InetSocketAddress(host, port));
		} catch (AlreadyConnectedException e) {
			throw new ConnectionException("Transport already connected", e);
		} catch (UnresolvedAddressException e) {
			throw new AddressException("Unresolved address: " + host, e);
		} catch (UnsupportedAddressTypeException e) {
			throw new AddressException("Unsupported address type", e);
		} catch (IOException e) {
			throw new OdetteFTPException("Unexcepted error: " + e.getMessage(),
					e);
		}

		return connected;
	}

	/**
	 * Terminate the ongoing connection. Once the connection is ended, it's not
	 * avaiable for further communication activity.
	 * 
	 * @throws OdetteFTPException
	 *             unexpected error
	 * @see org.fossilec.odettej.transport.Transport#disconnect()
	 */
	public synchronized void disconnect() throws OdetteFTPException {
		SocketChannel channel = getChannel();
		try {
			channel.close();
		} catch (IOException e) {
			throw new OdetteFTPException("unexpected error: " + e.getMessage(),
					e);
		}
	}

	/**
	 * Send a sequence of bytes through this transport implementation, using a
	 * TCP/IP socket channel, from the given buffer.
	 * <p>
	 * An attempt is made to send up to <i>r</i> bytes to the channel, where
	 * <i>r</i> is the number of bytes remaining in the buffer, that is,
	 * <code>dst.remaining()</code>, at the moment this method is invoked.
	 * 
	 * @param src
	 *            The buffer from which bytes are to be retrieved
	 * @return The number of bytes sent, possibly zero
	 * @throws NotConnectedException
	 *             Transport not yet connected
	 * @throws OdetteFTPException
	 *             unexpected exception
	 * @see org.fossilec.odettej.transport.Transport#send(ByteBuffer)
	 */
	public synchronized int send(ByteBuffer buffer) throws OdetteFTPException {
		int sentOctects = 0;
		SocketChannel channel = getChannel();

		try {
			sentOctects = channel.write(buffer);
		} catch (NotYetConnectedException e) {
			throw new NotConnectedException("Transport not yet connected", e);
		} catch (IOException e) {
			throw new OdetteFTPException("unexpected error: " + e.getMessage(),
					e);
		}

		return sentOctects;
	}

	/**
	 * Sent a sequence of bytes through this transport implementation, using a
	 * TCP/IP socket channel, from a subsequence of the given buffers.
	 * 
	 * @param srcs
	 *            The sequence of buffers from which bytes are to be retrieved
	 * @return The number of overall bytes sent, possibly zero
	 * @throws NotConnectedException
	 *             Transport not yet connected
	 * @throws OdetteFTPException
	 *             unexpected exception
	 * @see org.fossilec.odettej.transport.Transport#send(ByteBuffer[])
	 */
	public synchronized int send(ByteBuffer[] buffers)
			throws OdetteFTPException {
		int sentOctects = 0;
		SocketChannel channel = getChannel();

		try {
			sentOctects = (int) channel.write(buffers);
		} catch (NotYetConnectedException e) {
			throw new NotConnectedException("Transport not yet connected", e);
		} catch (IOException e) {
			throw new OdetteFTPException("unexpected error: " + e.getMessage(),
					e);
		}

		return sentOctects;
	}

	/**
	 * Receive a sequence of bytes from the TCP socket channel.
	 * <p>
	 * An attempt is made to receive up to <i>r</i> bytes from the channel,
	 * where <i>r</i> is the number of bytes remaining in the buffer, that is,
	 * <code>dst.remaining()</code>, at the moment this method is invoked.
	 * 
	 * @param dst
	 *            The buffer into which bytes are to be transferred
	 * @return The number of bytes received, possibly zero, or <code>-1</code>
	 *         if the channel has reached end-of-stream
	 * @throws NotConnectedException
	 *             Transport not yet connected
	 * @throws OdetteFTPException
	 *             Thrown by unexpected exception
	 * @see org.fossilec.odettej.transport.Transport#receive(ByteBuffer)
	 */
	public synchronized int receive(ByteBuffer dest) throws OdetteFTPException {
		int receivedOctects = 0;
		SocketChannel channel = getChannel();

		try {
			receivedOctects = channel.read(dest);
		} catch (NotYetConnectedException e) {
			throw new NotConnectedException("Transport not yet connected", e);
		} catch (IOException e) {
			throw new OdetteFTPException("unexpected error: " + e.getMessage(),
					e);
		}

		return receivedOctects;
	}

	/**
	 * Tells whether the underlying transport is connected.
	 * 
	 * @return <code>true</code> if this transport implementation is
	 *         connected.
	 * @throws OdetteFTPException
	 * @see org.fossilec.odettej.transport.Transport#isConnected()
	 */
	public boolean isConnected() throws OdetteFTPException {
		SocketChannel channel = getChannel();
		return channel.isConnected();
	}

	/**
	 * Returns local address used to fulfill Odette FTP session variables. On
	 * the <i>Initiator</i> side, it compose the "<code>odette.calling-address
	 * </code>"
	 * property. The <i>Responder</i> Odette FTP entity use this method to fill
	 * the "<code>odette.called-address</code>" property.
	 * 
	 * @return String representing the local address of the underlying network
	 *         protocol, as expected by session variables.
	 * @throws OdetteFTPException
	 * @see org.fossilec.odettej.transport.Transport#getLocalAddress()
	 */
	public String getLocalAddress() throws OdetteFTPException {
		SocketChannel channel = getChannel();
		return channel.socket().getLocalSocketAddress().toString();
	}

	/**
	 * Returns remote address used to fulfill Odette FTP session variables. On
	 * the <i>Initiator</i> side, it's used in the "<code>odette.called-address
	 * </code>".
	 * The <i>Responder</i> Odette FTP entity use this method to fill the "<code>odette.calling-address</code>"
	 * property.
	 * 
	 * @return String representing the remote address of the underlying network
	 *         protocol, as expected by session variables.
	 * @throws OdetteFTPException
	 * @see org.fossilec.odettej.transport.Transport#getRemoteAddress()
	 */
	public String getRemoteAddress() throws OdetteFTPException {
		SocketChannel channel = getChannel();
		return channel.socket().getRemoteSocketAddress().toString();
	}
}